//
//  AASFileManager.h
//
//  Created by andreas on Fri Nov 16 2001.
//  OS X 10.5 or later required.
/* Copyright (c) 2001-2009, Andreas Schwarz
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 
 Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 Neither the names of the Andreas Schwarz or immaterial nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */




//  Known issues: subscribing/getting file objects for a directory is not thread-safe. Need to add some locks and stuff.


#import <Foundation/Foundation.h>
#import <CoreServices/CoreServices.h>
#import "AASQueuedObjectOperation.h"





#pragma mark Error Values


// Error domain for NSError objects
extern NSString *AASFilesystemManagerErrorDomain;


/* Errors for AASFilesystemManagerErrorDomain */
typedef enum {
	AASErrorUnknown						= -99,	// Unknown error.
	AASErrorUserCanceled				= -1,	// User cancelled operation.
	AASErrorNonexistentFile				= -2,	// Source file doesn't exist.
	AASErrorNonexistentDestination		= -3,	// Destination folder doesn't exist.
	AASErrorNotAFolder					= -4,	// Expected a folder but wasn't a folder.
	AASErrorInvalidFile					= -5,	// File is invalid (nonspecific error, couldn't access the file for some unknown reason).
	AASErrorNoPermission				= -6,	// The user does not have permission to perform the requested operation on the file.
	AASErrorFileAlreadyExists			= -7,	// You are trying to move, copy, rename, or create a new file where a file with the same name already exists. It is unknown whether you have permission to overwrite the file.
	AASErrorFileNeedsOverwrite			= -8,	// Same as AASErrorFileAlreadyExists except it means you have permission to overwrite the file.
	AASErrorCannotOverwrite				= -9,	// Same as AASErrorFileAlreadyExists except it means you don't have permission to overwrite the file.
	AASErrorRecursiveMove				= -10,	// You are trying to move a directory into one of its own descendants.
	AASErrorRecursiveReplace			= -11	// You are trying to replace a directory with one of its own descendants.
} AASFMError;

extern NSString *AASErrorSourceItemKey;			// Key for the AASFileItem that represents the source file involved in the error
extern NSString *AASErrorDestinationItemKey;	// Key for the AASFileItem that represents the destination folder involved in the error
				// NSFilePathErrorKey			// NSFilePathErrorKey will contain the "destination" path in cases like rename operations where the final file does not exist yet.



#pragma mark -
#pragma mark File Operations Keys


/* File manipulation types */
typedef enum {
	AASCopyOperation	= 1,					// Denotes a file copy/duplicate operation
	AASMoveOperation	= 2,					// Denotes a file move operation
	AASTrashOperation	= 3,					// Denotes a move to trash operation
	AASAliasOperation	= 4,					// Denotes a make alias operation.
	AASLinkOperation	= 5,					// Denotes a make symlink operation
	AASRenameOperation	= 6						// Denotes a file rename operation. Note that if you pass a bunch of files for renaming, the preparation method will NOT check if your new filenames will run into each other, only if they will run into a preexisting file.
} AASFileOperationType;


/* File manipulation status values */
typedef enum {
	AASPreparingOperation	= 1,				// Denotes an operation which is waiting to begin
	AASPreflightOperation	= 2,				// Denotes preflight checking which occurs before some operations
	AASPerformingOperation	= 3,				// Denotes operation is in progress
	AASCompletedOperation	= 4,				// Denotes an operation which is complete
	AASCanceledOperation	= 5					// Denotes an operation which was canceled
} AASFileOperationStatus;


/* Keys for file manipulation dictionaries */
extern NSString *AASOpSourceItemKey;			// Key for the AASBasicInfoFileItem for the source file of a file operation (the file you intend to move, copy, etc.)
extern NSString *AASOpDestinationItemKey;		// Key for the AASBasicInfoFileItem for the destination folder of a file operation (where you intend to move the source to, etc.)
extern NSString *AASOpNewNameKey;				// Key for the NSString that will be the new name of a file in a rename.
extern NSString *AASOpSErrorKey;				// After preparation, some source files may have been tagged with an error using this key.
extern NSString *AASOpDErrorKey;				// After preparation, some destination files may have been tagged with an error using this key.
extern NSString *AASOpResolutionKey;			// The solution the sender presents for an error is set with this key.


/* Values for file manipulation resolutions. You can only return one resolution so it must satisfy both source and destination issue if you get both for a file. */
typedef enum {
	AASOpResolutionIgnore		= 0,			// This value might be useful for sender's internal operations but should not be set (the performOperation: method will ignore it).
	AASOpResolutionStop			= 1,			// This value might be useful for sender's internal operations, but rather than set this value (the performOperation: method will ignore it) the sender should instead just not call performOperation:.
	AASOpResolutionSkip			= 2,			// Skip this file (do not perform operation).
	AASOpResolutionOverwrite	= 3				// Overwrite the preexisting file with the new one you're moving.
} AASOpResolution;



/* Keys for dictionary that is returned to the specified sender and selector during a file operation */
//			AASOpSourceItemKey					// Source file item for file that is being processed or just finished. The actual file being worked on may be a child item if moving or copying.
//			AASOpDestinationItemKey				// Destination file item for file that  is being processed or just finished. For a rename operation, destination is nil;
extern NSString *AASOpCurrentPathKey;			// Path of current file being operated on. May be children of the source during a move or copy, otherwise is just the source.

extern NSString *AASOpContextInfoKey;			// Context info object supplied by sender when operation began
extern NSString *AASOpTypeKey;					// NSNumber containing AASFileOperationType value for this operation
extern NSString *AASOpErrorKey;					// If not nil, there was an error with the file and an NSError object is supplied.
extern NSString *AASOpCurrentStatusKey;			// Returns the current AASFileOperationStatusType type as NSNumber to report status.

extern NSString *AASOpRemainingBytesKey;		// Number of bytes remaining to copy.
extern NSString *AASOpCompletedBytesKey;		// Number of bytes completed so far.
extern NSString *AASOpTotalBytesKey;			// Number bytes total in the operation, or during preflight, number of bytes found so far.

extern NSString *AASOpTotalFileCountKey;		// Number of files being operated on total, or during preflight, number of files found so far.
extern NSString *AASOpCompletedFileCountKey;	// Number of files completed so far.
extern NSString *AASOpRemainingFileCountKey;	// Number of files left to operate on.

extern NSString *AASOpPercentageCompleteKey;	// Percentage complete.

extern NSString *AASOpDateStartedKey;			// NSDate object created when the operation began (useful for calculating remaining time).
extern NSString *AASOpDateProcessingStartedKey;	// NSDate object created when the processing (afterpreflight) portion of the operation began (useful for calculating remaining time).

extern NSString *AASOpOperationIdentifierKey;	// The operation identifier, same as returned by the initial call to performOperation:...




#pragma	mark -
#pragma mark File Tracking Protocols


@protocol AASFilesystemSubscriptionProtocol;

/* AASFileManager contains methods for tracking changes to files on disk. This is one of the most useful functions of the AASFilesystemTools framework. Tracking and updates is accomplished using FSEvents to get notified of changes to subscribed directories, then scanning each directory for changes. All the subscriber has to do is sit back and wait for updates! */








#pragma	mark -
#pragma mark File Manager Interface Declaration


@class AASDirectorySubscription;
@class AASFileItem;

@interface AASFileManager : NSObject {
	FSEventStreamRef stream;
	NSTimer *streamResetTimer;
	FSEventStreamEventId lastEvent;
	BOOL ignoreFurtherEvents;
	BOOL needsHeavyReset;
	NSMutableDictionary *subscribedPaths;
	NSMutableDictionary *dirtyFolders;
	
	NSOperationQueue *operationQueue;
	NSOperationQueue *utilityQueue;
	    
    NSMutableArray *removedDirectoriesReleaseTimers;
	
	NSMutableDictionary *fileOperations;
    
    Class fileClass;
    BOOL asynchronous;
    
    NSTimeInterval updateInterval;
    BOOL paused;
	
	unsigned short checkID;
	BOOL doingCheck;
	
	NSLock *resourceForkLock;
}



#pragma	mark -
#pragma mark File Manager Methods


/* Returns the AASFileManager instance (AASFileManager is a singleton). Before you use it you must set certain values, such as the file class (file class specifies what class you want the AASFileManager and AASDirectorySubscription objects to use for tracking files - AASFileItem [the default] or a subclass).  Filesystem tracking doesn't actually begin until there is at least one subscription. */
+ (AASFileManager *)defaultFileManager;


/* Call this as your program exits to gracefully clean up. Do not attempt to use the AASFileManager instance after this. */
- (void)shutdown;



/* Returns the class of file objects the AASFileManager and its AASDirectorySubscription instances are using. Default is AASFileItem. */
/* Subclassing the AASFileItem class can be useful if you want to have more information about each file in one convenient "central" location; see AASBasicInfoFileItem as an example.  There is only one AASFileItem object per file in the file system. */
/* The property must be set before any subscriber asks to track a directory or any method is called that creates an object based on the file class. If you do this correctly, the methods will return YES. If you attempt to set them afterward, they will return NO. */
@property (assign) Class fileClass;


/* The operation queue used for loading files. Has a very high maximum concurrent operation count. */
@property(readonly) NSOperationQueue * operationQueue;

/* The operation queue used various file operations. Uses a maximum concurrent operation count of 4 * number of processors on the machine. */
@property(readonly) NSOperationQueue * utilityQueue;

/* Operations that need to manipulate resource forks use this lock areound any resource fork manipulations to ensure threaded operations function properly; resource fork functions are not thread-safe on their own. If you app manipulates resource forks, use this lock to ensure no file operations trample on your data. */
@property(readonly) NSLock * resourceForkLock;





#pragma	mark -
#pragma mark File Operations Methods

/* These methods are for getting file objects for a directory as well as methods to manipulate files (move, copy, etc.) associated with file objects. */



/* Convenience methods */

/* Method for getting an autoreleased array of file objects that represent the contents of a directory without using an AASDirectorySubscription. If there's another object that has subscribed to the directory, this will be able to efficiently reuse that same collection. */
- (NSArray *)filesItemsForDirectory:(AASFileItem *)directory;

/* Takes an array of AASFileItem objects and makes an array of dictionaries suitable for passing to the file manipulation functions. Pass nil to destination for no destination (only for moving files to trash) or pass an AASFileItem for the directory you want to move/copy/alias your files to. If you need to rename files or send your files to a number of different destinations, build the dictionaries manually rather than using this method. */
- (NSArray *)fileDictionariesForFileItems:(NSArray *)fileItems withDestination:(AASFileItem *)destination;






/* File Manipulation */

/* Creates a new folder in the specified directory. If name is nil, folder name will be "untitled folder" (or "untitled folder 1" or some numbered variant as necessary). Returns a file item for the new folder. */
- (id)createNewFolderInDirectory:(AASFileItem *)directory withName:(NSString *)name sender:(id)sender error:(NSError **)error;


/* Takes an array of NSDictionaries defining the AASFileItems on which the operation will operate (using AASOpSourceItemKey and AASOpDestinationItemKey), and/or an NSString representing the file's new name (for rename operations only at this point. Someday we will implement move-with-new-name and copy-with-new-name but it isn't high priority). Returns an array of NSMutableDictionaries that include the same objects and keys, plus keys for any potential issues with the source or destination (with AASOpSIssueKey and AASOpDIssueKey), the values of which will be an NSNumber for the AASOpIssue. If issues exist, the sender should query the user or otherwise choose a course of action for that file, and then should add one of the resolution values (such as AASOpResolutionSkip or AASOpResolutionOverwrite) with the AASOpResolutionKey. Then the array may be passed to the performOperation: method */
- (NSMutableArray *)prepareForOperation:(AASFileOperationType)operation withFiles:(NSArray *)fileDicts sender:(id)sender;


/* Takes an array or dictionaries as prepared by prepareForOperation: and performs the defined operation on them. This method will return an NSString that uniquely identifies the operation (receiver may retain this to use with the cancelOperation: method, and it can be helprul to use as a key to keep track of multiple operations), or returns nil if the operation couldn't begin (for example, if every file was marked "skipped"). Operations are asynchronous and updates are returned to sender via the given update selector. Selector should take one argument, which will be an NSDictionary. */
- (NSString *)performOperation:(AASFileOperationType)operation withFiles:(NSArray *)fileDicts sender:(id)sender updateSelector:(SEL)uSel contextInfo:(id)info;

/* To cancel an operation, pass the operation id string into this method. Sender will receive one final  AASOpCompletedStatus status update. */
- (void)cancelOperation:(NSString *)operationID;



/* Returns an NSError object for the specified type and file items that has a localized description and appropriate recovery options */
- (NSError *)errorForType:(AASFMError)error underlyingError:(NSError *)uError source:(AASFileItem *)source dest:(AASFileItem *)dest destPath:(NSString *)destPath;



#pragma	mark -
#pragma mark File Tracking Methods


/* The following property must be set before any subscriber asks to track a directory (for example, in your app delegate's applicationWillFinishLaunching: method). */
/* Sets whether or not to read files from the filesystem in separate threads. Subscribers will be notified of changes on the main thread (see the AASDirectorySubscription header for more). Default is NO. */
@property(assign, getter=isAsynchronous) BOOL asynchronous;

/* Set the minimum interval at which you want to recieve updates. Set this value before any subscriber asks to track a directory. Default is 2.0. */
@property(assign) NSTimeInterval updateInterval;

/* Pauses or unpauses filesystem tracking by the AASFileManager.  While it is paused, no updating goes on.  Default is NO (but no updating occurs until the AASFileManager has at least one subscriber). */
@property(assign, getter=isPaused) BOOL paused;



/* Subscribing and unsubscribing to file tracking services. */


/* Subscribes a particular object, adding the directory represented by "file" to the list of directories to check.  Returns the AASDirectorySubscription object that the subscriber subscribed to, or nil if it doesn't succeed.  The subscriber should keep a reference to the AASDirectorySubscription object as it will be needed when unsubscribing. */
- (AASDirectorySubscription *)subscribe:(id <AASFilesystemSubscriptionProtocol>)subscriber forFile:(AASFileItem *)file;
/* Same as subscribe: forFile: but can be passed a path instead of a file object. */
- (AASDirectorySubscription *)subscribe:(id <AASFilesystemSubscriptionProtocol>)subscriber forPath:(NSString *)path;
/* Same as subscribe: forFile: but can be passed an NSURL instead of a file object. */
- (AASDirectorySubscription *)subscribe:(id <AASFilesystemSubscriptionProtocol>)subscriber forURL:(NSURL *)url;

/* Unsubscribes a subscriber from the subscription given. */
- (void)unsubscribe:(id <AASFilesystemSubscriptionProtocol>)unsubscriber forSubscription:(AASDirectorySubscription *)sub;

/* Checks all the AASDirectorySubscription instances to see if there's already an instance for this file object.  Returns it if it exists, returns a newly created instance otherwise. */
- (AASDirectorySubscription *)subscriptionForFile:(AASFileItem *)file;



/* Checking the filesystem */

/* You may call this to check filesystem for updates if for some reason you don't want to wait for the automatic update notification. */
- (void)checkFilesystem:(id)sender;

/* Lets you notify AASFileManager there has been a change in a particular folder. FSEvents might take one to five seconds to notice a change, so if you need a faster update you can report changes via this method, then call -checkFilesystem:. */
- (void)changeHappenedInFolder:(AASFileItem *)folderItem;

@end

















