//
//  AASFilesystemSubscription.h
//  AASFilesystemTools
//
//  Created by Andreas Schwarz on 11/23/09.
//  Copyright 2009 immaterial. All rights reserved.
//

#import <Cocoa/Cocoa.h>


/* Generic superclass for various subscription classes, not meant to be used directly. */





#pragma mark Protocols

/* Subscribers must implement this method.  It is called whenever the watched files change. It gives an array of dictionaries containing information about the changed files.  The keys to these dictionaries are below. When using an AASDirectorySubscription subclass set to load asynchronously, this is called periodically while the files load, giving AASFileAppeared events with the newly loaded files.  To check if an AASDirectorySubscription is currently loading, use the isLoading method.  After all loading has finished, the AASDirectorySubscription will check if any changes happened during loading, and if so will call this with the appropriate changes, or will call it with an empty changes array (so that if the subscriber needs to check isLoading each time the method is called to see when the loading finishes, it will have a chance to get NO from isLoading at an appropriate time). */

@class AASFilesystemSubscription;

@protocol AASFilesystemSubscriptionProtocol <NSObject>
- (void)subscription:(AASFilesystemSubscription *)subscription foundChanges:(NSArray *)changes;
@end



#pragma mark -
#pragma mark Keys

////////////// Dictionary key strings ////////////////

/* Key for the file event string
 The file event constants (directly below) identify what sort of file event occurred for the file.  The file event key will return an NSNumber object with one of these values. */
extern NSString *AASFileEventKey;

typedef enum {
	AASFileMovedEvent			= 1,	// The file moved our of its original directory (may also be renamed).
	
	AASFileRenamedEvent			= 2,	// The file was renamed but remains in its original the directory.
	
	AASFileRemovedEvent			= 4,	// The file was deleted or is otherwise untrackable. The file should return NO for -isValid.
	
	AASFileUpdatedEvent			= 8,	// File data or metadata updated, but no moves or renames.
	
	AASFileAppearedEvent		= 16,	// The file appeared in the parent directory (newly created or just moved in).
	
	AASFileMovedInParentEvent	= 32	// The file "moved" (its path changed) but it is still within its parent directory. This happens when the parent or a grandparent directory moved. When you get this notification it is possible, however unlikely, that the file has also been renamed.
} AASFileEvents;


/* Key for the file object for this event
 The file object is of whatever class was specified to the default AASFileManager instance. */
extern NSString *AASFileKey;

/* If the file replaced a file (in case of AASFileAppearedEvent) or file was replaced by a file (in case of AASFileRemovedEvent) then this key will also be in the event dictionary. The value will be the file object for the replaced file (in case of AASFileAppearedEvent) or the replacing file (in case of AASFileRemovedEvent). */
extern NSString *AASFileReplacementKey;





@interface AASFilesystemSubscription : NSObject {
    NSMapTable *subscribers;
    Class fileClass;
}



#pragma mark -
#pragma mark Subscribers


/* Adds a subscriber.  The subscriber is not retained, so that the subscriber may unsubscribe itself in its dealloc method.  For obvious reasons, it is important the subscriber make sure it is unsubscribed before or during its deallocation. Do not subscribe to the same directory multiple times with the same subscriber. */
- (void)subscribe:(id <AASFilesystemSubscriptionProtocol>)subscriber;

/* Removes a subscriber (if it exists), returns the number of subscribers left. ALWAYS unsubscribe before releasing the subscriber object (you can unsubscribe in the subscriber's dealloc method). Only removes the subscriber once, so don't subscribe multiple times to one AASDirectorySubscription with the same subscriber. */
- (int)unsubscribe:(id <AASFilesystemSubscriptionProtocol>)subscriber;

/* Notifies subscribers of the specified changes. Typically called by subclasses, don't call this yourself. */
- (void)notifySubscribers:(NSArray *)changes;


#pragma mark -
#pragma mark Subscription Information


/* Returns the number of subscribers the reciever has. */
@property(readonly) NSUInteger subscriberCount;



#pragma mark -
#pragma mark Properties and Methods for Subclasses

/* Returns nil. */
@property(readonly) NSString * path;

/* Returns nil. */
@property(readonly, retain) NSURL * url;

/* Always returns YES. */
@property(readonly) BOOL isValid;

/* Always returns NO. */
@property(readonly) BOOL isLoading;

/* Returns nil. */
@property(readonly) NSArray * files;

/* Returns nil. */
@property(readonly) id directoryFile;

/* Does nothing. */
- (void)loadImmediately;


@end
